use std::{
    path::PathBuf,
    sync::{Arc, Mutex},
};

use anyhow::anyhow;
use clap::Command;

use crate::{actions, context::CuddleContext, model::CuddleScript};

#[derive(Debug, Clone)]
#[allow(dead_code)]
struct CuddleAction {
    script: CuddleScript,
    path: PathBuf,
    name: String,
}
#[allow(dead_code)]
impl CuddleAction {
    pub fn new(script: CuddleScript, path: PathBuf, name: String) -> Self {
        Self { script, path, name }
    }

    pub fn execute(self) {
        match self.script {
            CuddleScript::Shell(_s) => {
                match actions::shell::ShellAction::new(
                    self.name.clone(),
                    format!(
                        "{}/scripts/{}.sh",
                        self.path
                            .to_str()
                            .expect("action doesn't have a name, this should never happen"),
                        self.name
                    ),
                )
                .execute()
                {
                    Ok(()) => {}
                    Err(e) => {
                        log::error!("{}", e)
                    }
                }
            }
            CuddleScript::Dagger(_d) => {}
        }
    }
}

#[derive(Debug, Clone)]
pub struct CuddleCli<'a> {
    scripts: Vec<CuddleAction>,
    context: Arc<Mutex<Vec<CuddleContext>>>,
    command: Option<Command<'a>>,
}

impl<'a> CuddleCli<'a> {
    pub fn new(context: Arc<Mutex<Vec<CuddleContext>>>) -> anyhow::Result<CuddleCli<'a>> {
        let mut cli = CuddleCli {
            scripts: vec![],
            context: context.clone(),
            command: None,
        };

        cli = cli.process_scripts().build_cli();

        Ok(cli)
    }

    fn process_scripts(mut self) -> Self {
        if let Ok(context_iter) = self.context.clone().lock() {
            for ctx in context_iter.iter() {
                if let Some(scripts) = ctx.plan.scripts.clone() {
                    for (name, script) in scripts {
                        self.scripts
                            .push(CuddleAction::new(script.clone(), ctx.path.clone(), name))
                    }
                }
            }
        }

        self
    }

    fn build_cli(mut self) -> Self {
        let mut root_cmd = Command::new("cuddle")
            .version("1.0")
            .author("kjuulh <contact@kasperhermansen.com>")
            .about("cuddle is your domain specific organization tool. It enabled widespread sharing through repositories, as well as collaborating while maintaining speed and integrity")
            .propagate_version(true);

        let mut execute_cmd = Command::new("x").about("x is your entry into your domains scripts, scripts inherited from parents will also be present here");

        for script in self.scripts.iter() {
            let action_cmd = Command::new(script.name.clone());

            // TODO: Some way to add an about for clap, requires conversion from String -> &str

            execute_cmd = execute_cmd.subcommand(action_cmd);
        }

        root_cmd = root_cmd.subcommand(execute_cmd);

        self.command = Some(root_cmd);

        self
    }

    pub fn execute(self) -> anyhow::Result<Self> {
        if let Some(mut cli) = self.command.clone() {
            let matches = cli.clone().get_matches();

            let res = match matches.subcommand() {
                Some(("x", exe_submatch)) => {
                    log::trace!("executing x");

                    match exe_submatch.subcommand() {
                        Some((name, _action_matches)) => {
                            log::trace!(action=name; "running action");
                            match self.scripts.iter().find(|ele| ele.name == name) {
                                Some(script) => {
                                    script.clone().execute();
                                    Ok(())
                                }
                                _ => (Err(anyhow!("could not find a match"))),
                            }
                        }
                        _ => (Err(anyhow!("could not find a match"))),
                    }
                }
                _ => (Err(anyhow!("could not find a match"))),
            };

            match res {
                Ok(()) => {}
                Err(e) => {
                    let _ = cli.print_long_help();
                    return Err(e);
                }
            }
        }

        Ok(self)
    }
}
